#ifndef _TRON_H
#define _TRON_H

class function
{
public:
	virtual double fun(double *w, long long &data_access) = 0 ;
	virtual void grad(double *w, double *g, long long &data_access) = 0 ;
	virtual void Hv(double *s, double *Hs, long long &data_access) = 0 ;
	virtual void HvXv(double *s, double *Hs, double *Xs, long long &data_access) = 0;
	virtual double sample_Hv(double *s, double *Hs, long long &data_access) = 0 ;
	virtual void Xv(double *v, double *Xv, long long &data_access) = 0;

	virtual void fix_a_sample(int sample_size) = 0;
	virtual void get_matrix(double *XP, double *inner, double *A, int limit, int num, int idx) = 0;
	virtual double fun_alpha(double *Xs, double *d, double *w, double alpha, double wTd, double dTd) = 0;
	virtual void get_alpha(double *alpha, double *beta, double gTs, double gTsold, double sTs, double sTsold, double soldTsold, double *Xs, double *Xsold, bool second_direction, long long &num_data) = 0;
	virtual double line_search(double *d, double *w, double *g, double *Xs, double dTd, double wTd, double gTd, double &f, long long &data_access) = 0 ;
	virtual void step(double dTd, double wTd, double alpha, double &f) = 0 ;

	virtual int get_nr_variable(void) = 0 ;
	virtual int get_nr_instance(void) = 0 ;
	virtual size_t get_nr_element(void) = 0 ;
	virtual void get_diag_preconditioner(double *M, long long &data_access) = 0 ;
	virtual void get_grad_and_diagonal(double *w, double *g, double *M, long long &data_access) = 0;
	virtual void get_sample_diag_preconditioner(double *M, long long &data_access) = 0 ;
	virtual void setup_use(int *M, long long &data_access) = 0 ;
	virtual void get_use(int *M, long long &data_access) = 0 ;
	virtual ~function(void){}
};

class TRON
{
public:
	TRON(const function *fun_obj, double eps = 0.1, double eps_cg = 0.1, int max_iter = 1000, int num_directions = -1, int subsample_ratio = 1, int max_cg_iter = int(1e3), double alpha_pcg = 0, int seed = 2019);
	~TRON();

	void tron(double *w);
	void set_print_string(void (*i_print) (const char *buf));

private:
	int trpcg(double *g, double *M, double *s, double *w, double *r, double delta, long long &data_access);
	double norm_inf(int n, double *x);

	double eps;
	double eps_cg;
	int max_iter;
	int num_directions;
	int subsample_ratio;
	int max_cg_iter;
	double alpha_pcg;
	function *fun_obj;
	void info(const char *fmt,...);
	void (*tron_print_string)(const char *buf);
};
#endif
